#!/bin/python3
#%%
import time
t_init = time.time()

import hashlib
import json
import os
import datetime
import pickle
import multiprocessing

#%%
import pandas as pd
import numpy as np

#%%
from tools.std_lib.basic import json_dumps, get_timestamp, human_readable_size

#%%
from tools.std_lib.log import ColorfulLogger, apply_color
C = apply_color
Cp = lambda p: apply_color(p, color='underline,blue')
log = ColorfulLogger(basename='basic_data_tools.py')

#%%

def get_size_speed_str(
        file: bytes, 
        time_spent: float = None,
        new_line = True
        ):
    if isinstance(file, str):
        if os.path.isfile(file):
            file_name = file
            file_size = os.path.getsize(file)
        else:
            log.error(f'file not found')
            raise ValueError(f'file not found')
    elif isinstance(file, bytes):
        file_name = '(bytes)'
        file_size = len(file)
    file_size_str = human_readable_size(file_size=file_size)
    file_name = apply_color(file_name, 'underline,blue')
    if time_spent is None:
        string = f'{file_size_str}    {file_name}'
    else:
        if time_spent > 0:
            speed_str = human_readable_size(file_size / time_spent)
        elif time_spent == 0:
            speed_str = f'inf'
        else:
            log.error(f'time_spent < 0', func = 'get_size_speed()')
        string = f'{file_size_str}    {file_name}\n'
        string += f'Time = {time_spent:.03f}s, Speed = {speed_str}/s'
    if new_line:
        string = '\n' + string
    return string


def pickle_load(pickle_path: str)->None:
    t0 = time.time()
    log.info(f'pickle_load("{Cp(pickle_path)}")')
    log.info(f'{get_size_speed_str(pickle_path)}')
    var = pickle.load(open(pickle_path, 'rb'))
    t1 = time.time()
    log.info(f'{get_size_speed_str(pickle_path, t1 - t0)}')
    return var

silence_pickle_load = lambda pickle_path: pickle.load(open(pickle_path, 'rb'))


def pickle_dump(var: object, pickle_path = None, name = 'var', makedirs = False) -> None:
    t0 = time.time()
    if pickle_path is None:
        pickle_path = f'./tmp/pickle_dump/{get_timestamp()}-{var.__class__.__name__}.pickle'
    print(f'\npickle_path = \'{pickle_path}\'')
    print(f'{name} = pickle_load(pickle_path = pickle_path)\n')
    if makedirs:
        os.makedirs(os.path.dirname(pickle_path), exist_ok=True)
    if hasattr(var, 'info'):
        var.info()
    log.info(f'pickle_dump(var = {type(var)}, pickle_path = "{Cp(pickle_path)}")')
    if 'to_pickle' in dir(var):
        var.to_pickle(pickle_path)
    else:
        pickle.dump(var, open(pickle_path, 'wb'))
    t1 = time.time()
    log.info(get_size_speed_str(pickle_path, t1 - t0))
    return pickle_path

silence_pickle_dump = lambda var, pickle_path: pickle.dump(var, open(pickle_path, 'wb'))


def multiprocessing_pool_map_list(func, iterable, process = None):
    if process is None or process == 'auto':
        process = int(multiprocessing.cpu_count() * .9)
    log.info(f'process = {process}', func = 'multiprocessing_pool_map_list()')
    pool = multiprocessing.Pool(processes=process)
    result = pool.map(func=func, iterable=iterable)
    pool.close()
    pool.join()
    return result


def multiprocessing_pool_map_dict(func, iterable, process = None):
    if isinstance(iterable, list):
        result = multiprocessing_pool_map_list(func=func, iterable=iterable, process=process)
    elif isinstance(iterable, dict):
        keys = list(iterable.keys())
        values = [iterable[k] for k in keys]
        result_values = multiprocessing_pool_map_list(func=func, iterable=values, process=process)
        result = dict(zip(keys, result_values))
    else:
        raise ValueError(f'iterable is not list or dict')
    return result


def normal_map_dict(func, iterable, process = None):
    if isinstance(iterable, list):
        result = [func(k) for k in iterable]
    elif isinstance(iterable, dict):
        result = {k: func(v) for k, v in iterable.items()}
    else:
        raise ValueError(f'iterable is not list or dict')
    return result


def multiprocessing_pool_map(
        func, 
        iterable, 
        process = None, 
        disable_multiprocessing = False,
        to_series = False,
        to_frame = False,
        json_normalize = False,
        ):
    print()
    func_name = 'multiprocessing_pool_map()'
    log.info('started', func = func_name)
    columns = None
    if disable_multiprocessing:
        log.warning(f'disable_multiprocessing = {disable_multiprocessing}', func = func_name)
    if isinstance(iterable, range):
        iterable = list(iterable)
    if isinstance(iterable, pd.Series):
        index = iterable.index
        iterable = {idx: iterable[idx] for idx in iterable.index}
        to_series = True
    elif isinstance(iterable, pd.DataFrame):
        columns = iterable.columns
        iterable = {idx: iterable[idx] for idx in iterable.columns}
        # to_frame = True

    if len(iterable) == 1:
        disable_multiprocessing = True
        log.warning(f'len(iterable) = {len(iterable)} and set disable_multiprocessing = {disable_multiprocessing}', func = func_name)

    t0 = time.time()
    if not disable_multiprocessing:
        result = multiprocessing_pool_map_dict(
            func=func, 
            iterable=iterable, 
            process=process
            )
    else:
        result = normal_map_dict(
            func=func, 
            iterable=iterable, 
            process=process
            )
    t1 = time.time()
    time_spent = t1 - t0
    speed = f'{len(iterable) / time_spent:.03f}' if time_spent != 0 else 'inf'
    log.info(f'len(iterable)/time = {len(iterable)} / {t1 - t0:.03f}s = {speed} /s', func = func_name)
    
    if to_series:
        result = pd.Series(result, index=index)

    if to_frame:
        result = pd.DataFrame(result)
        if columns is not None:
            result.columns = columns

    if json_normalize:
        log.info(f'pd.json_normalize()', func = func_name)
        result = pd.json_normalize(result)
        result.index = index
        result = result[sorted(result.columns)]
        
    log.info('completed', func = func_name)
    print()
    return result


def save_df_to_csv(data_df: pd.DataFrame, base_path: str, makedirs = False):
    if makedirs:
        os.makedirs(os.path.dirname(base_path), exist_ok=True)
    save_path = base_path if base_path.endswith('.csv') else base_path + '.csv' 
    t0 = time.time()
    data_df.to_csv(save_path)
    t1 = time.time()
    log.info(get_size_speed_str(save_path, t1 - t0), func = 'save_df_to_csv()')
    return save_path

def save_df_to_excel(data_df: pd.DataFrame, base_path: str, makedirs = False):
    if makedirs:
        os.makedirs(os.path.dirname(base_path), exist_ok=True)
    save_path = base_path if base_path.endswith('.xlsx') else base_path + '.xlsx' 
    t0 = time.time()
    data_df.to_excel(save_path)
    t1 = time.time()
    log.info(get_size_speed_str(save_path, t1 - t0), func = 'save_df_to_excel()')
    return save_path

def save_df_to_pickle(data_df: pd.DataFrame, base_path: str, makedirs = False):
    if makedirs:
        os.makedirs(os.path.dirname(base_path), exist_ok=True)
    save_path = base_path if base_path.endswith('.pickle') else base_path + '.pickle' 
    t0 = time.time()
    data_df.to_pickle(save_path)
    t1 = time.time()
    log.info(get_size_speed_str(save_path, t1 - t0), func = 'save_df_to_pickle()')
    return save_path


def max_gap(numbers):
    if len(numbers) < 2:
        ret = float('nan')
    else:
        sorted_numbers = sorted(numbers)
        gaps = [sorted_numbers[i+1] - sorted_numbers[i] for i in range(len(sorted_numbers) - 1)]
        ret = max(gaps)
    return ret


def max_index_gap(df: pd.DataFrame, index_range = [], verbose = False):
    if isinstance(index_range, int):
        index_range = [1, index_range]
    nums = df.index.to_list()
    nums.extend(index_range)
    gap = max_gap(nums)
    if verbose:
        string = apply_color(f'max_index_gap = {gap}', 'green')
        print(string)
    return gap

def check_dataframe(data_df: pd.DataFrame, error = 'raise', index_range = []):
    if True and len(data_df) > 0:
        a = data_df.index[0]
        is_range = (data_df.index == range(a, a+len(data_df))).all()
    else:
        a = float('nan')
        is_range = False
    if is_range:
        string = apply_color(f'index == range({a}, len+{a})', 'green')
        print('info:', string)
    else:
        string = apply_color(f'index != range({a}, len+{a})', 'green')
        print('info:', string)
    print('index.dtype.name:', data_df.index.dtype.name)
    print('index.is_monotonic_increasing:', data_df.index.is_monotonic_increasing)
    print('index.is_unique:', data_df.index.is_unique)
    print('columns.dtype.name:', data_df.columns.dtype.name)
    print('columns.is_monotonic_increasing:', data_df.columns.is_monotonic_increasing)
    print('columns.is_unique:', data_df.columns.is_unique)
    dtypes = data_df.dtypes.value_counts()
    dtypes.index = [k.name for k in dtypes.index]
    dtypes = dtypes.sort_index()
    print('shape:', data_df.shape)
    print('dtypes:')
    print(dtypes)
    wrong_dtype = sorted(set(dtypes.index) - {'datetime64[ns]', 'Float64', 'float64', 'Int64', 'int64', 'Int32', 'int32', 'object', 'bool'})
    wrong_dtype_exists = len(wrong_dtype) > 0
    if wrong_dtype_exists:
        print(f'warning: wrong_dtype = {wrong_dtype}')
    bool_ls = [
        data_df.index.dtype.name in ['int64', 'Int64', 'int32', 'int64', 'Float64',],
        data_df.index.is_unique,
        data_df.columns.dtype.name == 'object',
        data_df.columns.is_unique,
        not wrong_dtype_exists,
    ]
    r = all(bool_ls)
    max_index_gap(data_df, index_range = index_range, verbose = True)
    print('DataFrame:', apply_color(r, 'green'))
    if error == 'raise' and not r:
        raise ValueError('check DataFrame.')
    if 'info' in dir(data_df):
        print('info():')
        print(data_df.info())
    return r



def save_dataframe(
        data_df: pd.DataFrame, 
        base_path: str, 
        formats = ['pickle'], 
        makedirs = False
        ):
    func_name = 'save_dataframe()'
    log.info('started', func = func_name)

    check_dataframe(data_df)

    if 'xlsx' not in formats:
        pickle_path = base_path
        if 'csv' in formats:
            csv_path = save_df_to_csv(data_df=data_df, base_path=base_path, makedirs=makedirs)
            pickle_path = csv_path + '.pickle'
        if 'pickle' in formats:
            save_df_to_pickle(data_df=data_df, base_path=pickle_path, makedirs=makedirs)

    elif 'csv' not in formats:
        pickle_path = base_path
        if 'xlsx' in formats:
            csv_path = save_df_to_excel(data_df=data_df, base_path=base_path, makedirs=makedirs)
            pickle_path = csv_path + '.pickle'
        if 'pickle' in formats:
            save_df_to_pickle(data_df=data_df, base_path=pickle_path, makedirs=makedirs)
    
    elif 'csv' in formats and 'xlsx' in formats:
        save_df_to_csv(data_df=data_df, base_path=base_path, makedirs=makedirs)
        save_df_to_excel(data_df=data_df, base_path=base_path, makedirs=makedirs)
        if 'pickle' in formats:
            save_df_to_pickle(data_df=data_df, base_path=base_path, makedirs=makedirs)

    log.info('completed', func = func_name)
    return None

    
    
def dict_to_plain(var: object):

    # basic
    if isinstance(var, list) or isinstance(var, tuple):
        var = [dict_to_plain(k) for k in var]
    elif isinstance(var, dict):
        var = {dict_to_plain(k): dict_to_plain(v) for k, v in var.items()}

    # pandas
    elif isinstance(var, pd.DataFrame):
        var = dict_to_plain(var.to_dict())
    elif isinstance(var, pd.Series):
        var = dict_to_plain(var.to_dict())
    elif isinstance(var, pd.Interval):
        var = [var.left, var.right]
    elif isinstance(var, pd.Index):
        var = dict_to_plain(var.to_list())

    # numpy
    elif isinstance(var, np.ndarray):
        var = var.tolist()
    elif any([
        isinstance(var, np.int64), 
        isinstance(var, np.int32),
        ]):
        var = int(var)
    elif any([
        isinstance(var, np.float64), 
        isinstance(var, np.float32)
        ]):
        var = float(var)
    elif isinstance(var, np.bool_):
        var = bool(var)

    # othoers
    elif callable(var):
        var = f'{var.__module__}.{var.__name__}'
    
    elif any([
        var is None,
        isinstance(var, str), 
        isinstance(var, int), 
        isinstance(var, float), 
        isinstance(var, bool),
        ]) :
        var = var
    
    else:
        raise ValueError(f'unkown type(var) = {type(var)}, var = {var}')

    return var


def get_hash(
        var: object, n = None, alg = 'sha256', salt = b''
        ):

    if var is None:
        var_bytes = b''
    elif isinstance(var, bytes):
        var_bytes = var
    else:
        var = dict_to_plain(var)
        if isinstance(var, dict) or isinstance(var, list):
            var_str = json.dumps(var, ensure_ascii=False, indent=4)
        elif var is None:
            var_str = ''
        else:
            var_str = str(var)
        var_bytes = var_str.encode('utf-8')

    if isinstance(salt, str):
        salt = salt.encode('utf-8')
    var_bytes = salt + var_bytes

    if alg == 'sha256':
        hash_instance = hashlib.sha256()
    elif alg == 'md5':
        hash_instance = hashlib.md5()
    hash_instance.update(var_bytes)
    hash_hex = hash_instance.hexdigest()

    if n is not None:
        hash_hex = hash_hex[:n]

    return hash_hex


def list_set_operation(lists, operation='intersection'):
    if not lists: 
        return []
    if len(lists) == 1:
        return sorted(lists[0])
    
    sets = [set(lst) for lst in lists]
    if operation == 'union':
        result = sets[0]
        for s in sets[1:]:
            result = result.union(s)
    else:
        result = sets[0]
        for s in sets[1:]:
            result = result.intersection(s)
    return list(result)


def split_dataframe(data_df: pd.DataFrame, n_part: int, n_split = 3, ):
    if n_split == 0:
        raise ValueError('n_part == 0')
    if n_part > n_split:
        raise ValueError('n_part > n_split')
    if n_split <= 0:
        raise ValueError('n_split <= 0')
    chunk_size = len(data_df) // n_split
    start = n_part * chunk_size
    end = (n_part+1) * chunk_size
    if n_split - n_part == 1: 
        end = len(data_df)
    part_df = data_df.iloc[start:end]
    return part_df


def get_monthly_stat_df(date_series):
    month_series = date_series.dt.strftime('%Y-%m')
    count = month_series.value_counts().sort_index()
    percent = count / count.sum()
    stat_df = pd.DataFrame({
        'count': count, 'percent': percent
    })
    return stat_df


def stat_month_df(data_df: pd.DataFrame):
    func_name = 'stat_month_df()'
    log.info('start', func = func_name)
    month_df = data_df.select_dtypes(include='datetime64[ns]')
    dfs = {}
    for col in month_df.columns:
        stat_df = get_monthly_stat_df(month_df[col])
        log.info(f'\ncol = "{col}"', stat_df.round(3))
        dfs[col] = stat_df
        print()
    log.info('completed', func = func_name)
    return dfs


#######################################################################




#%%

def save_and_reset_index(df: pd.DataFrame, origin_index_name = None):
    if origin_index_name is not None:
        df.index.name = origin_index_name
    df.insert(0, df.index.name, df.index)
    df.index = pd.Index(range(1, len(df)+1), name='index')
    return df


def get_color_prefix(n: object):
    if isinstance(n, int):
        prefix = f'\033[{n}m'
    elif isinstance(n, list):
        prefix = ''.join([f'\033[{k}m' for k in n])
    return prefix


def colorize_str(string: str, color = [91], show = False):
    string = str(string)
    if isinstance(color, int):
        color = [color]
    colorful_string = get_color_prefix(color) + string + '\033[0m'
    if show:
        print(colorful_string)
    return colorful_string


def t_print(*args, color = None):
    text = f'T={time.time() - t_init:1.0f}s'
    for var in args:
        if isinstance(var, pd.Series) or isinstance(var, pd.DataFrame):
            if text[-1] == ' ':
                text = text[:-1] + '\n'
            elif text[-1] != '\n':
                text = text + '\n'
            end = '\n'
        elif isinstance(var, list) or isinstance(var, dict) :
            if text[-1] == ' ':
                text = text[:-1] + '\n'
            elif text[-1] != '\n':
                text = text + '\n'
            end = '\n'
            var = json_dumps(dict_to_json(var))
            var = colorize_str(var, 92)
        elif isinstance(var, set) :
            if text[-1] == ' ':
                text = text[:-1] + '\n'
            elif text[-1] != '\n':
                text = text + '\n'
            end = '\n'
        else:
            if text[-1] != '\n':
                text = text + ' '
            end = ''
        text += str(var) + end
    while text[-1] in ['\n', ' ']:
        text = text[:-1]
    if color is not None:
        text = colorize_str(text, color=color)
    print(text)
    return None


def add_prefix_to_filename(file_path, prefix, sep = '-'):
    base_name = os.path.basename(file_path)
    new_file_name = prefix + sep + base_name
    new_path = os.path.join(os.path.dirname(file_path), new_file_name)
    print(f'new_path = \'{new_path}\'')
    return new_path


def add_suffix_to_filename(path: str, suffix = '', var = None, sep = '-'):
    p1, p2 = os.path.splitext(path)
    if len(suffix) > 0:
        suffix = sep + str(suffix)
    if var is not None:
        if isinstance(var, pd.DataFrame):
            suffix += f'{sep}shape_%d_%d'%var.shape
        elif isinstance(var, dict) or isinstance(var, list):
            suffix += f'{sep}n{len(var)}'
    new_path = p1 + suffix + p2
    print(f'new_path = \'{new_path}\'')
    return new_path


def dict_to_json(d, digit = None):
    r = {}
    if isinstance(d, str) or isinstance(d, int) or d is None:
        r = d
    elif type(d).__module__ == 'numpy' and type(d).__name__[:3] == 'int':
        r = dict_to_json(int(d), digit=digit)
    elif type(d).__module__ == 'numpy' and type(d).__name__[:5] == 'float':
        r = dict_to_json(float(d), digit=digit)
    elif isinstance(d, float):
        if str(d) == 'inf':
            r = 'inf'
        elif str(d) == '-inf':
            r = '-inf'
        elif str(d) == 'nan':
            r = 'nan'
        else:
            if digit is not None:
                d = round(d, digit)
            r = d
    elif type(d) == dict:
        r = {dict_to_json(k, digit = digit): dict_to_json(v, digit = digit) for k, v in d.items()}
    elif type(d) == list:
        r = [dict_to_json(k, digit = digit) for k in d]
    elif type(d) == tuple:
        r = [dict_to_json(k, digit=digit) for k in d]
    elif isinstance(d, pd.DataFrame):
        r = dict_to_json(d.to_dict(), digit = digit)
    elif isinstance(d, pd.Series):
        r = dict_to_json(d.to_dict(), digit = digit)
    elif isinstance(d, pd.Interval):
        r = dict_to_json(d.__str__(), digit = digit)
    elif type(d).__module__ == 'ray.tune.sample':
        r = d.domain_str
    else:
        r = f'{type(d).__module__}.{type(d).__name__}'
        if r not in [
            'builtins.function',
            'functools.partial',
            ]:
            print(f'warning: unknown type found: {r}')
    return r





def get_file_size(file_path, verbose = False, eta = None):
    if isinstance(file_path, bytes):
        file_size = len(file_path)
        file_path = colorize_str('bytes', color=[4, 94])
    elif isinstance(file_path, str):
        file_size = os.path.getsize(file_path)
        file_path = colorize_str(file_path, color=[4, 94])
    if verbose or eta is not None:
        print(f"File size: \n{human_readable_size(file_size=file_size)}    {file_path}")
        if eta is not None:
            if eta != 0:
                v = human_readable_size(file_size / eta)
            else:
                v = 'inf'
            print(f'rate = {v}/s, eta = {eta:.2f}s')
    return file_size

